/* Copyright 2012 Red Hat, Inc.
 *
 * This software may be freely redistributed under the terms of the GNU
 * public license.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <ctype.h>
#include <dirent.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/types.h>

#include "vmbus.h"
#include "modules.h"

#include "kudzuint.h"


static void vmbusFreeDevice(struct vmbusDevice *dev)
{
  if (dev->vmbusclass) free (dev->vmbusclass); dev->vmbusclass = NULL;
  freeDevice((struct device *) dev);
}


static void vmbusWriteDevice(FILE * file, struct vmbusDevice *dev)
{
  writeDevice(file, (struct device *) dev);
  fprintf(file, "vmbusclass: %s\n", dev->vmbusclass);
  fprintf(file, "devnum: %d\n", dev->devnum);
  fprintf(file, "reserved: %d\n", dev->reserved);
}


static int vmbusCompareDevice(struct vmbusDevice *dev1, struct vmbusDevice *dev2)
{
  int x,y;
  x = compareDevice((struct device *)dev1,(struct device *)dev2);
  if (x && x!=2) return x;

  /* cut & paste from compareDevice... we need to honor this */
  if (dev1->type == CLASS_NETWORK && dev2->type == CLASS_NETWORK &&
      dev1->classprivate && dev2->classprivate &&
      !strcmp((char *)dev1->classprivate,(char *)dev2->classprivate))
      return 0;

  if (dev1->vmbusclass && dev2->vmbusclass && (y = strcasecmp(dev1->vmbusclass, dev2->vmbusclass))) return y;
  if ((y = dev1->devnum - dev2->devnum)) return y;
  if ((y = dev1->reserved - dev2->reserved)) return y;

  return x;
}


struct vmbusDevice *vmbusNewDevice(struct vmbusDevice *old)
{
  struct vmbusDevice *ret;

  ret = malloc(sizeof(struct vmbusDevice));

  memset(ret, '\0', sizeof(struct vmbusDevice));
  ret = (struct vmbusDevice *) newDevice((struct device *) old, (struct device *) ret);

  ret->bus = BUS_VMBUS;
  ret->newDevice = vmbusNewDevice;
  ret->freeDevice = vmbusFreeDevice;
  ret->writeDevice = vmbusWriteDevice;
  ret->compareDevice = vmbusCompareDevice;
  if (old && old->bus == BUS_VMBUS) {
    ret->vmbusclass = strdup(old->vmbusclass);
    ret->devnum = old->devnum;
    ret->reserved = old->reserved;
  } else {
    ret->devnum = -1;
    ret->reserved = -1;
  }

  return ret;
}


void vmbusFreeDrivers() {
}


int vmbusReadDrivers(char *filename)
{
  aliases = readAliases(aliases, filename, "vmbus");
  return 0;
}


static enum deviceClass vmbusToKudzu(char * vmbusclass)
{
  if (!strcasecmp (vmbusclass, VMBUS_CLASS_IDE)) return CLASS_IDE;
  if (!strcasecmp (vmbusclass, VMBUS_CLASS_SCSI)) return CLASS_SCSI;
  if (!strcasecmp (vmbusclass, VMBUS_CLASS_HID)) return CLASS_MOUSE;
  if (!strcasecmp (vmbusclass, VMBUS_CLASS_NETWORK)) return CLASS_NETWORK;

  return CLASS_OTHER;
}


struct vmbusDevice *getVmbusDevice(char *name, struct vmbusDevice *ret, enum deviceClass probeClass)
{
  struct vmbusDevice *tmp;
  struct dirent *entry;
  char *tname, *tname2;
  char *modalias, *drv;
  DIR *d;
  int cwd;

  d = opendir(name);
  if (!d) return ret;

  cwd = open(".", O_RDONLY);
  chdir(name);
  tmp = vmbusNewDevice(NULL);
  tname = name;

  /* vmbus device name has the following format : vmbus_{reserved}_{devnum} */
  /* example : vmbus_0_1 */

  /* seek underscore and go to the next char*/
  while (*(tname++) != '_');

  /* read the reserved value & check the endptr */
  tmp->reserved = strtoul(tname, &tname2, 10); 
  if (*(tname2++) != '_') return ret; /* wrong device name format */

  /* read the device number & check the endptr */
  tmp->devnum = strtoul(tname2, &tname, 10); 
  if (*tname != '\0') return ret; /* wrong device name format */

  /* taking values from files in the device directory */
  while ((entry = readdir(d))) {

    if (entry->d_name[0] == '.') continue;

    if (!strcmp (entry->d_name, "class_id")) {
      if (tmp->vmbusclass) free (tmp->vmbusclass);
      tmp->vmbusclass = __readString(entry->d_name);
    }

    if (!strcmp (entry->d_name, "modalias")) {
      modalias = __readString(entry->d_name);
      if (modalias) {
	drv = aliasSearch(aliases, "vmbus", modalias + 6);
	if (drv) {
	  if (tmp->driver) free (tmp->driver);
          tmp->driver = strdup(drv);
	}
      }
      free(modalias);
    }

    if (!strncmp(entry->d_name, "net:", 4)) {
      __getSysfsDevice((struct device *)tmp, ".", "net:", 0);
      __getNetworkAddr((struct device *)tmp, tmp->device);
    }

  }

  /* shouldn't happen at all */
  if (!tmp->vmbusclass) tmp->vmbusclass = strdup("{}");

  /* convert the vmbuss class to kudzu class */
  tmp->type = vmbusToKudzu(tmp->vmbusclass);

  if (probeClass & tmp->type) {  
    /* it's a wanted class, process it */

    if (tmp->desc) free(tmp->desc);

    switch (tmp->type) {

      case CLASS_NETWORK:
        if (!tmp->device) tmp->device = strdup("eth");
	tmp->desc = strdup(DESC_VMBUS_NETWORK);
        break;

      case CLASS_IDE:
	tmp->desc = strdup(DESC_VMBUS_IDE);
	break;
                
      case CLASS_SCSI:
	tmp->desc = strdup(DESC_VMBUS_SCSI);
	break;

      case CLASS_MOUSE:
	tmp->desc = strdup(DESC_VMBUS_HID);
	break;

      case CLASS_OTHER:
	if ( !strcmp (tmp->vmbusclass, VMBUS_CLASS_UTIL_1) ||
	     !strcmp (tmp->vmbusclass, VMBUS_CLASS_UTIL_2) ||
	     !strcmp (tmp->vmbusclass, VMBUS_CLASS_UTIL_3) ||
	     !strcmp (tmp->vmbusclass, VMBUS_CLASS_UTIL_4) ) 
        {
	  tmp->desc = strdup(DESC_VMBUS_UTIL);
	} else {
	  tmp->desc = strdup(DESC_VMBUS_OTHER);
	}
	break;

      default:
	tmp->desc = strdup(DESC_VMBUS_OTHER);
    } 

    /* append to device list */
    if (ret) tmp->next = (struct device *)ret;
    ret = tmp;
  } else {  
    /* class not wanted, free the device */
    vmbusFreeDevice(tmp);
  }
 
  /* do some cleaning */
  closedir(d);
  fchdir(cwd);
  close(cwd);

  return ret;
}


struct device *vmbusProbe(enum deviceClass probeClass, int probeFlags, struct device *devlist)
{
  int init_list = 0;
  int cwd;

  if ((probeClass & CLASS_OTHER) || (probeClass & CLASS_SCSI)  ||
      (probeClass & CLASS_IDE)   || (probeClass & CLASS_MOUSE) ||
      (probeClass & CLASS_NETWORK))
  {
    DIR *dir;
    struct dirent *entry;
		
    if (!getAliases (aliases, "vmbus")) {
      vmbusReadDrivers(NULL);
      init_list = 1;
    }

    /* check if "/sys/bus/vmbus/devices" exists */
    dir = opendir("/sys/bus/vmbus/devices");
    if (!dir) goto out;

    /* process all device entries in "/sys/bus/vmbus/devices" */
    cwd = open(".",O_RDONLY);
    chdir("/sys/bus/vmbus/devices");
    while ((entry = readdir(dir))) {
      if (!strncmp (entry->d_name, "vmbus_", 6)) {
	devlist = (struct device *) getVmbusDevice (entry->d_name, (struct vmbusDevice *)devlist, probeClass);
      }
    }

    /* do some cleaning */
    closedir(dir);
    fchdir(cwd);
    close(cwd);
  }

out:
  if (init_list) vmbusFreeDrivers();

  return devlist;
}
